from json import load as json_load, dump as json_dump
from pathlib import Path

PARENT = Path(__file__).parent
all_locators_path = PARENT / "all_secret_locators.json"
with open(all_locators_path, "r") as f:
    all_locator_dicts = json_load(f)

curated_path = PARENT / "curated.json"
generic_path = PARENT / "generic.json"
cloud_path = PARENT / "cloud.json"
endpoint_path = PARENT / "endpoint.json"
backend_path = PARENT / "backend.json"
secret_path = PARENT / "secret.json"
key_locators_path = PARENT / "key_locators.json"
aws_path = PARENT / "aws.json"
azure_path = PARENT / "azure.json"
gcp_path = PARENT / "gcp.json"

def matches_terms(locator_dicts, terms):
    matching = []
    for locator_dict in locator_dicts:
        if any(term.lower() in value.lower() for term in terms for value in map(str, locator_dict.values())):
            matching.append(locator_dict)
    return matching

generic_locators = matches_terms(all_locator_dicts, ["generic"])
cloud_locators = matches_terms(all_locator_dicts, ["cloud", 'aws', 'azure', 'gcp', 'amazon', 'google', 'microsoft'])
endpoint_locators = matches_terms(all_locator_dicts, ["endpoint", "domain", 'host', 'address', 'interface', 'rest', 'http'])
key_locators = matches_terms(all_locator_dicts, ["key", "token"])
secret_locators = matches_terms(all_locator_dicts, ["secret"])
curated_locators = matches_terms(all_locator_dicts, ["rsa", "ssh", "secret", "openai", "aws", "azure", "gcp", "amazon", "google", "microsoft", "endpoint", "domain", 'host', 'address', 'interface', 'rest', 'http',])
aws_locators = matches_terms(all_locator_dicts, ["aws", "amazon"])
azure_locators = matches_terms(all_locator_dicts, ["azure", "microsoft"])
gcp_locators = matches_terms(all_locator_dicts, ["gcp", "google"])

for locator_dicts, path in zip(
    [generic_locators, cloud_locators, endpoint_locators,
     key_locators, curated_locators, secret_locators,
     aws_locators, azure_locators, gcp_locators,
     ],

     [generic_path, cloud_path, endpoint_path, key_locators_path,
      curated_path, secret_path,
      aws_path, azure_path, gcp_path,]):

    print(f"Writing {len(locator_dicts)} locators to {path}.")
    with path.open("w+") as f:
        json_dump(locator_dicts, f, indent=4)



# © 2023 Lucas Faudman.
# Licensed under the MIT License (see LICENSE for details).
# For commercial use, see LICENSE for additional terms.
from pathlib import Path
from typing import Optional, Generator, Iterable, Literal
from datetime import datetime
from yaml import dump as yaml_dump
from json import dump as json_dump

from .decompiler import Decompiler
from .secret_scanner import SecretScanner, SecretResult

class APKScanner:
    def __init__(self,
                 decompiler_kwargs: dict,
                 scanner_kwargs: dict,
                 output_file: Optional[Path] = None,
                 output_format: str = "json",
                 groupby: Literal["file", "locator", "both"] = "both",
                 cleanup: bool = True,
                 ):
        # workers
        self.decompiler = Decompiler(**decompiler_kwargs)
        self.secret_scanner = SecretScanner(**scanner_kwargs)
        # output
        self.output_file = output_file or Path(f"./secrets_output.{output_format}")
        self.output_format = output_format
        self.groupby = groupby
        self.cleanup = cleanup
        self.output_written = False
        self.cleaned_up = False
        # state tracking
        self.decompiling = {}
        self.decompilers_count_by_ext = {}
        self.decompile_start_time = None
        self.decompile_elapsed_time = None
        self.scanning = set()
        self.scan_start_time = None
        self.scan_elapsed_time = None
        self.last_scanned = None
        self.last_secret = None
        # counters
        self.num_files = 0
        self.num_decompiled = 0
        self.num_decompile_success = 0
        self.num_decompile_errors = 0
        self.num_scanning = 0
        self.num_scanned = 0
        self.num_secrets = 0
        self.num_unique_secrets = 0
        # results
        self.decompile_results = {}
        self.secrets_results = []
        self.unique_secrets = set()

        print(f"\nInitialized Decompiler:\n- {self.decompiler}")
        print(f"\nInitialized Secret Scanner:\n- {self.secret_scanner}")
        print(f"\nDecompiler Binaries:\n- " + "\n- ".join(map(str, self.decompiler.binary_paths.values())))
        print(f"\nSecret Locator Files:\n- " + "\n- ".join(map(str, self.secret_scanner.secret_locator_files)))
        print(f"\nOutput File:\n- {self.output_file.absolute()}\n")

    def print_status(self, end='\r'):
        if self.decompiling and not self.scanning:
            status = "Decompiling"
        elif self.decompiling and self.scanning:
            status = "Decompiling and Scanning"
        elif not self.decompiling and self.scanning:
            status = "Scanning"
        else:
            status = "COMPLETE"
        status_message = f"Status: {status} | "
        if self.num_files:
            status_message += f"Decompiled: {self.num_decompiled}/{self.num_files} ({self.num_decompiled/self.num_files*100:3f}%) | "
        if self.num_scanning:
            status_message += f"Scanned: {self.num_scanned}/{self.num_scanning} ({self.num_scanned/self.num_scanning*100:3f}%) | "
        if self.num_secrets:
            status_message += f"Secrets: {self.num_secrets} ({self.num_unique_secrets} unique) | "
        print(status_message, end=end, flush=True)

    def print_secret_found(self, secret_result: SecretResult) -> None:
        print(f"Found {secret_result.locator.name}: \033[92m{secret_result.secret[:100]}\033[0m in {secret_result.file_path}:{secret_result.line_number} (line {secret_result.line_number})\n")

    def files_to_decompile_generator(self, file_paths: Iterable[Path]) -> Generator[Path, None, None]:
        for file_path in file_paths:
            self.num_files += 1
            ext = file_path.suffix
            if not (num_decompilers := self.decompilers_count_by_ext.get(ext)):
                num_decompilers = self.decompiler.num_binaries_to_run_on_ext(ext)
                self.decompilers_count_by_ext[ext] = num_decompilers
            self.decompiling[file_path.stem] = num_decompilers

            if not self.decompile_start_time:
                self.decompile_start_time = datetime.now()
                print(f"\nDecompiling started at {self.decompile_start_time.strftime('%H:%M:%S:%SS')}\n")

            self.print_status()
            yield file_path

    def decompiled_files_generator(self, file_paths: Iterable[Path]) -> Generator[Path, None, None]:
        for file_path, output_dir, decompiled_files, success in self.decompiler.decompile_concurrently(file_paths):
            self.decompile_results[output_dir] = (file_path, decompiled_files, success)
            self.decompiling[file_path.stem] -= 1

            if success and decompiled_files:
                self.num_decompile_success += 1

                if not self.scan_start_time:
                    self.scan_start_time = datetime.now()
                    print(f"\nScanning started at {self.scan_start_time.strftime('%H:%M:%S:%SS')}\n")

                for decompiled_file in decompiled_files:
                    self.num_scanning += 1
                    self.scanning.add(decompiled_file)
                    self.print_status()
                    yield decompiled_file

            else:
                self.num_decompile_errors += 1

            if self.decompiling[file_path.stem] == 0:
                self.num_decompiled += 1
                del self.decompiling[file_path.stem]
                self.print_status('\n')

        self.decompiler.concurrent_executor.shutdown()
        if self.decompile_start_time:
            self.decompile_elapsed_time = datetime.now() - self.decompile_start_time
            print(f"\nDecompiling COMPLETE. Decompiled {self.num_decompiled} files with {self.num_decompile_errors} errors. Elapsed time: {self.decompile_elapsed_time}\n")


    def scan_secret_results_generator(self, file_paths: Generator[Path, None, None]) -> Generator[SecretResult, None, None]:
        self.scan_start_time = None
        for file_path, file_secret_results in self.secret_scanner.scan_concurrently(file_paths):
            if file_path in self.scanning:
                self.scanning.remove(file_path)
                self.num_scanned += 1

            self.last_scanned = file_path
            self.print_status()
            yield from file_secret_results

        self.print_status('\n')
        if self.scan_start_time:
            self.scan_elapsed_time = datetime.now() - self.scan_start_time
            print(f"\nScanning COMPLETE. Scanned {self.num_scanned} files with {self.num_secrets} secrets found. Elapsed time: {self.scan_elapsed_time}\n")

    def decompile_and_scan(self, file_paths: Iterable[Path]) -> list[SecretResult]:
        self.decompile_and_scan_start_time = datetime.now()
        num_secrets_before = self.num_secrets
        num_unique_secrets_before = self.num_unique_secrets

        files_to_decompile = self.files_to_decompile_generator(file_paths)
        decompiled_files = self.decompiled_files_generator(files_to_decompile)
        for secret_result in self.scan_secret_results_generator(decompiled_files):
            self.num_secrets += 1
            self.secrets_results.append(secret_result)
            if secret_result.secret not in self.unique_secrets:
                self.num_unique_secrets += 1
                self.print_secret_found(secret_result)
                self.unique_secrets.add(secret_result.secret)
            self.print_status()

        self.total_elapsed_time = datetime.now() - self.decompile_and_scan_start_time
        print(f"Found \033[92m{self.num_secrets - num_secrets_before} new secrets\033[0m ({self.num_unique_secrets - num_unique_secrets_before} unique).")
        print(f"Decompiled {self.num_decompiled} files with {self.num_decompile_errors} errors in {self.decompile_elapsed_time}.")
        print(f"Scanned {self.num_scanned} files and found {self.num_secrets} secrets in {self.scan_elapsed_time}.")
        print(f"Total Elapsed time: {datetime.now() - self.decompile_and_scan_start_time}")

        return self.secrets_results

    def make_secret_result_serializable(self, secret_result: SecretResult) -> dict:
        secret_result_dict = {
            "secret": secret_result.secret,
            "file_path": str(secret_result.file_path),
            "line_number": secret_result.line_number,
            "locator": secret_result.locator.name,
        }
        try:
            secret_result_dict["secret"] = secret_result_dict["secret"].decode()
        except Exception as e:
            print(f"Error decoding secret: {e}")
            secret_result_dict["secret"] = f'{secret_result_dict["secret"]}'
        return secret_result_dict

    def group_results_by_locator(self) -> dict[str, list[SecretResult]]:
        results_by_locator = {}
        for secret_result in self.secrets_results:
            serializable_secret_result = self.make_secret_result_serializable(secret_result)
            results_by_locator.setdefault(secret_result.locator.id, []).append(serializable_secret_result)
        return results_by_locator

    def group_results_by_input_file(self) -> dict[str, list[SecretResult]]:
        results_by_input_file = {}
        for secret_result in self.secrets_results:
            for output_dir, (file_path, decompiled_files, success) in self.decompile_results.items():
                # TODO maybe group by output_dir (decompilier) ?
                if secret_result.file_path in decompiled_files:
                    serializable_secret_result = self.make_secret_result_serializable(secret_result)
                    results_by_input_file.setdefault(str(file_path), []).append(serializable_secret_result)
                    break
        return results_by_input_file

    def write_output(self):
        print(f"\nWriting output to {self.output_file}", end="\r")
        if self.groupby == "file":
            results = self.group_results_by_input_file()
        elif self.groupby == "locator":
            results = self.group_results_by_locator()
        elif self.groupby == "both":
            results = {
                "by_file": self.group_results_by_input_file(),
                "by_locator": self.group_results_by_locator()
            }

        with self.output_file.open("w") as f:
            if self.output_format == "json":
                json_dump(results, f, indent=4)
            elif self.output_format == "yaml":
                yaml_dump(results, f, default_flow_style=False)
            else:
                f.write(str(results))

        self.output_written = True
        print(f"Output written to {self.output_file}")


    def do_cleanup(self, **cleanup_concurrency_kwargs):
        if self.cleanup and not self.cleaned_up:
            self.decompiler.cleanup(**cleanup_concurrency_kwargs)
            self.decompiler.concurrent_executor.shutdown(wait=False, cancel_pending=True)
            self.secret_scanner.concurrent_executor.shutdown(wait=False, cancel_pending=True)
        self.cleaned_up = True


    def __del__(self):
        if getattr(self, "secrets_results", False) and not getattr(self, "output_written", False):
            self.write_output()

        if getattr(self, "cleanup", False) and not getattr(self, "cleaned_up", False):
            self.do_cleanup(concurrency_type=None)



# © 2023 Lucas Faudman.
# Licensed under the MIT License (see LICENSE for details).
# For commercial use, see LICENSE for additional terms.
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor, as_completed, Future
from typing import Callable, Iterable, Iterator, Generator, Literal, Optional, TypeVar

T = TypeVar("T")

# Defined outside of class to allow for standalone use
def execute_concurrently(
    func: Callable[..., T],
    *iterables: Iterable,
    concurrency_type: Optional[Literal["thread", "process", "main", False]] = "thread",
    results_order: Literal["completed", "submitted"] = "completed",
    max_workers: Optional[int] = None,
    chunksize: int = 1,
    timeout: Optional[int] = None,
    shutdown: bool = True,
    wait: bool = True,
    cancel_pending: bool = False,
    executor: Optional[ThreadPoolExecutor | ProcessPoolExecutor] = None,
    **executor_init_kwargs,
) -> Generator[T, None, Optional[ThreadPoolExecutor | ProcessPoolExecutor]]:
    """Execute function concurrently with arguments from iterables.

    Args:
        func: Function to execute concurrently that takes as many arguments as there are iterables.
        *iterables: Iterables of args to pass to function calls in parallel e.g. ([call0_arg0, call1_arg0], [call0_arg1, call1_arg1], ...)
        concurrency_type: Concurrency type used to execute function. (Multithreading, Multiprocessing, or Main/Single Thread) Defaults to "thread".
        results_order: Order to yield results. Either in order completed or order submitted. Defaults to "completed".
        max_workers: Maximum number of threads or processes to use. Defaults to None (number of CPUs).
        chunksize: The size of the chunks the iterable will be broken into before being passed to a child process. Only used when concurrency_type is "process". Defaults to 1.
        timeout: The maximum number of seconds to wait. If None, then there is no limit on the wait time.
        shutdown: Whether to shutdown executor after completion or to return it for reuse. Defaults to True. If False, returns executor instance on StopIteration.
        wait: Whether to wait for executor to shutdown. Defaults to True.
        cancel_pending: Whether to cancel pending futures on shutdown. Defaults to False.
        executor: Reuse an existing ThreadPoolExecutor or ProcessPoolExecutor instance. Defaults to None.
        executor_init_kwargs: Additional keyword arguments to pass to ThreadPoolExecutor or ProcessPoolExecutor constructor. Defaults to None.

    Yields:
        Results from function execution with arguments from iterables in order of completion or submission.
    """

    # Single-threaded execution in main thread (concurrency_type = "main"|False|None)
    if not concurrency_type or concurrency_type == "main":
        for args in zip(*iterables):
            yield func(*args)
        return

    # Multi-threaded or multi-process execution (concurrency_type = "thread"|"process")
    if not executor:
        executor_cls = ProcessPoolExecutor if "proc" in concurrency_type else ThreadPoolExecutor
        executor = executor_cls(max_workers=max_workers, **executor_init_kwargs)

    if "submit" in results_order:
        # Yield results in order of submission
        yield from executor.map(func, *iterables, timeout=timeout, chunksize=chunksize)
    else:
        # Yield results in order of completion
        futures_generator = (executor.submit(func, *args) for args in zip(*iterables))
        yield from map(Future.result, as_completed(futures_generator, timeout=timeout))

    # Shutdown executor or return it on StopIteration to be reused
    if shutdown:
        return executor.shutdown(wait=wait, cancel_futures=cancel_pending)
    else:
        return executor


class ConcurrentExecutor:
    def __init__(
        self,
        concurrency_type: Optional[Literal["thread", "process", "main", False]] = "thread",
        results_order: Literal["completed", "submitted"] = "completed",
        max_workers: Optional[int] = None,
        chunksize: int = 1,
        timeout: Optional[int] = None,
        shutdown: bool = True,
        wait: bool = True,
        cancel_pending: bool = False,
        executor: Optional[ThreadPoolExecutor | ProcessPoolExecutor] = None,
        **executor_init_kwargs,
    ) -> None:
        self.concurrency_type = concurrency_type
        self.results_order = results_order
        self.max_workers = max_workers
        self.chunksize = chunksize
        self.timeout = timeout
        self._shutdown = shutdown
        self.wait = wait
        self.cancel_pending = cancel_pending
        self.executor = executor
        self.executor_init_kwargs = executor_init_kwargs

    def map(self, func: Callable[..., T], *iterables: Iterable, **kwargs) -> Iterator[T]:
        self.executor = yield from execute_concurrently(
            func,
            *iterables,
            **{
                "concurrency_type": self.concurrency_type,
                "results_order": self.results_order,
                "max_workers": self.max_workers,
                "chunksize": self.chunksize,
                "shutdown": self._shutdown,
                "timeout": self.timeout,
                "wait": self.wait,
                "cancel_pending": self.cancel_pending,
                "executor": self.executor,
                **self.executor_init_kwargs,
                **kwargs,
            },
        )

    def shutdown(self, wait: Optional[bool] = None, cancel_pending: Optional[bool] = None) -> None:
        if self.executor:
            self.executor.shutdown(
                wait=wait if wait is not None else self.wait,
                cancel_futures=cancel_pending if cancel_pending is not None else self.cancel_pending
            )
            self.executor = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.shutdown()

    def __del__(self):
        self.shutdown()

    def __repr__(self) -> str:
        return f"CuncurrentExecutor(concurrency_type={self.concurrency_type}, results_order={self.results_order}, max_workers={self.max_workers}, chunksize={self.chunksize}, timeout={self.timeout}, shutdown={self._shutdown}, wait={self.wait}, cancel_pending={self.cancel_pending}, executor={self.executor}, executor_init_kwargs={self.executor_init_kwargs})"



# © 2023 Lucas Faudman.
# Licensed under the MIT License (see LICENSE for details).
# For commercial use, see LICENSE for additional terms.
from subprocess import run, DEVNULL, SubprocessError
from pathlib import Path
from shutil import which, rmtree
from os import access, X_OK
from shlex import split as shlex_split
from typing import Optional, Iterator, Iterable, Literal

from enjarify import enjarify
from .concurrent_executor import ConcurrentExecutor

class Decompiler:
    CONFIG = {
        "jadx": {
            "binary": which("jadx") or "/usr/local/bin/jadx",
            "output_arg": "--output-dir",
            "deobf_arg": "--deobf",
            "extra_args": [],
            "file_exts": {".apk", ".xapk", ".jar", ".dex", ".class", ".smali", ".zip", ".aar", ".arsc", ".aab", ".jadx.kts"}
        },
        "apktool": {
            "binary": which("apktool") or "/usr/local/bin/apktool",
            "output_arg": "--output",
            "deobf_arg": "--force-manifest",
            "extra_args": ["d", "--force", "--keep-broken-res"],
            "file_exts": {".apk", ".xapk"}
        },
        "procyon": {
            "binary": which("procyon-decompiler") or "/usr/local/bin/procyon-decompiler",
            "output_arg": "-o",
            "deobf_arg": "-renames",
            "extra_args": [],
            "file_exts": {".jar", ".dex", ".class"}
        },
        "cfr": {
            "binary": which("cfr-decompiler") or "/usr/local/bin/cfr-decompiler",
            "output_arg": "--outputdir",
            "deobf_arg": "--antiobf",
            "extra_args": [],
            "file_exts": {".jar", "dex", "class"}
        },
        "krakatau": {
            "binary": which("krakatau") or "/usr/local/bin/krakatau",
            "output_arg": "--out",
            "deobf_arg": "",
            "extra_args": ["dis"],
            "file_exts": {".jar", ".zip", ".class"}
        },
        "fernflower": {
            "binary": which("fernflower") or "/usr/local/bin/fernflower",
            "output_arg": "",
            "deobf_arg": "",
            "extra_args": [],
            "file_exts": {".jar", ".class"}
        }
    }

    def __init__(
        self,
        binaries: Optional[dict[str, Optional[Path|str]]|Iterable[str]] = None,
        enjarify_choice: Literal["auto", "never", "always"] = "auto",
        deobfuscate: bool = False,
        extra_args: Optional[list[str]] = None,
        output_suffix: str = "-decompiled",
        working_dir: Path = Path("/tmp/apk-secret-scanner"),
        overwrite: bool = False,
        remove_failed_output_dirs: bool = True,
        suppress_output: bool = False,
        **concurrent_executor_kwargs,
    ):
        self.binary_paths = self.validate_binary_paths(binaries)
        self.enjarify = self.validate_enjarify_choice(enjarify_choice)
        self.deobfuscate = deobfuscate
        self.extra_args = self.validate_extra_args(extra_args)
        self.output_suffix = output_suffix
        self.working_dir = working_dir
        self.overwrite = overwrite
        self.remove_failed_output_dirs = remove_failed_output_dirs
        self.suppress_output = suppress_output
        self.concurrent_executor = ConcurrentExecutor(**{"concurrency_type": "thread", **concurrent_executor_kwargs})
        self.output_dirs = {}

    def validate_binary_paths(self, binaries: Optional[dict[str, Optional[Path|str]]|Iterable[str]]) -> dict[str, Path]:
        if not binaries:
            # Use JADX binary by default if no binaries specified
            jadx_path = self.CONFIG["jadx"]["binary"]
            print(f"No decompiler binaries specified. Defaulting to JADX binary at {jadx_path}.")
            binaries = {"jadx": jadx_path}

        if not isinstance(binaries, dict):
            # Handle any iterable of binary names
            binaries = dict.fromkeys(binaries)

        binary_paths = {}
        for binary_name, binary_path in binaries.items():
            binary_path =  Path(binary_path or self.CONFIG[binary_name]["binary"])
            if not binary_path.exists():
                print(f"Skipping {binary_name}. Binary not found: {binary_path}")
                print(f"Use --{binary_name} <PATH> to specify the path to the binary.")
                continue
            elif not access(binary_path, X_OK):
                print(f"Skipping {binary_name}. Binary not executable: {binary_path}")
                print(f"Use chmod +x {binary_path} to make it executable.")
                continue

            print(f"Found {binary_name} binary: {binary_path}")
            binary_paths[binary_name] = binary_path

        if not binary_paths:
            raise FileNotFoundError("No valid decompiler binaries found.")

        return binary_paths

    def num_binaries_to_run_on_ext(self, file_ext: str) -> int:
        return sum(1 for binary_name in self.binary_paths if file_ext in self.CONFIG[binary_name]["file_exts"])

    def validate_enjarify_choice(self, enjarify_choice: Literal["auto", "never", "always"]) -> bool:
        enjarify_needed = set(self.binary_paths) - {"jadx", "apktool"}
        if enjarify_choice == "auto":
            enjarify = bool(enjarify_needed)
        else:
            enjarify = enjarify_choice == "always"

        if enjarify_needed and not enjarify:
            raise ValueError(f"Enjarify is needed for {enjarify_needed} but enjarify_choice is {enjarify_choice}.")

        return enjarify

    def validate_extra_args(self, extra_args: Optional[list[str]]) -> dict[str, list[str]]:
        extra_args_dict = {binary_name: self.CONFIG[binary_name]["extra_args"] for binary_name in self.binary_paths}
        if not extra_args:
            return extra_args_dict
        for extra_arg in extra_args:
            binary_name, *args = shlex_split(extra_arg)
            if binary_name not in self.binary_paths:
                print(f"Skipping extra args for {binary_name}. Binary not found.")
                continue
            extra_args_dict[binary_name] = args
        return extra_args_dict


    def make_args(self, binary_name: str, file_path: Path, output_path: Path) -> list[str]:
        args = [self.binary_paths[binary_name], *self.extra_args.get(binary_name, ()), self.CONFIG[binary_name]["output_arg"], output_path]
        if self.deobfuscate and (deobf_arg := self.CONFIG[binary_name].get("deobf_arg")) and deobf_arg not in args:
            args.append(deobf_arg)
        args.append(file_path)
        return list(map(str, args))


    def try_run_binary(self, binary_name: str, file_path: Path, output_path: Path) -> bool:
        args = self.make_args(binary_name, file_path, output_path)
        kwargs = {"stdout": DEVNULL, "stderr": DEVNULL} if self.suppress_output else {}
        kwargs["check"] = False
        try:
            print(f"Running {binary_name} on {file_path.name}")
            result = run(args, **kwargs) # type: ignore
            return True
        except SubprocessError as e:
            print(f"Error Running {binary_name} on {file_path.name}: {e}")
            return False

    def get_output_dir(self, file_path: Path) -> Path:
        output_dir = self.working_dir / (file_path.stem + self.output_suffix)
        self.output_dirs[file_path.stem] = output_dir
        if not output_dir.exists():
            print(f"Creating output directory: {output_dir}")
            output_dir.mkdir(parents=True, exist_ok=True)
            print(f"Output directory created: {output_dir}")
        return output_dir

    def enjarify_file(self, file_path: Path) -> Path:
        if file_path.suffix not in {".apk", ".dex"}:
            print(f"Skipping {file_path.name}. Enjarify only works on .apk and .dex files.")
            return file_path

        jar_file = (self.get_output_dir(file_path) / file_path.stem).with_suffix(".jar")
        if jar_file.exists() and not self.overwrite:
            return jar_file

        try:
            print(f"\nEnjarifying {file_path.name} to {jar_file.name}")
            enjarify(file_path, jar_file, overwrite=True, quiet=self.suppress_output)
            print(f"Successfully enjarified {file_path.name} to {jar_file}")
        except Exception as e:
            print(f"Error enjarifying {file_path.name}: {e}")
            jar_file.unlink(missing_ok=True)

        return jar_file

    def decompile(self, binary_name_file_path: tuple[str, Path]) -> tuple[Path, Path, Optional[set[Path]], bool]:
        binary_name, file_path = binary_name_file_path
        output_dir = self.get_output_dir(file_path) / binary_name
        if output_dir.exists() and not self.overwrite:
            success = True
        else:
            output_dir.mkdir(parents=True, exist_ok=True)
            success = self.try_run_binary(binary_name, file_path, output_dir)

        if success:
            print(f"Successfully decompiled {file_path.name} with {binary_name}")
        elif self.remove_failed_output_dirs:
            print(f"Erorr decompiling {file_path.name} with {binary_name}.")
            self.remove_output_dir(output_dir)

        if success or not self.remove_failed_output_dirs:
            print(f"\nIndexing decompiled files in {output_dir}...")
            decompiled_files = set((*filter(Path.is_file, output_dir.rglob("*")),))
            print(f"Found {len(decompiled_files)} decompiled files for {file_path.name}")
        else:
            decompiled_files = None

        return file_path, output_dir, decompiled_files, success

    def enjarify_concurrently(self, file_paths: Iterable[Path]) -> Iterator[Path]:
        yield from self.concurrent_executor.map(self.enjarify_file, file_paths)

    def binary_name_file_path_generator(self, file_paths: Iterable[Path]) -> Iterator[tuple[str, Path]]:
        file_paths_generator = self.enjarify_concurrently(file_paths) if self.enjarify else file_paths
        for file_path in file_paths_generator:
            for binary_name in self.binary_paths:
                if file_path.suffix in self.CONFIG[binary_name]["file_exts"]:
                    yield binary_name, file_path

    def decompile_concurrently(self, file_paths: Iterable[Path]) -> Iterator[tuple[Path, Path, Optional[set[Path]], bool]]:
        yield from self.concurrent_executor.map(self.decompile, self.binary_name_file_path_generator(file_paths))

    def remove_output_dir(self, output_dir: Path) -> Path:
        if output_dir.exists() and output_dir.is_dir():
            print(f"Removing: {output_dir}")
            rmtree(output_dir)
        return output_dir

    def cleanup(self, **concurrency_kwargs):
        output_dirs = list(self.output_dirs.values())
        print(f"\nRemoving {len(output_dirs)} decompiled output directories...")
        for output_dir in self.concurrent_executor.map(
            self.remove_output_dir, output_dirs, **concurrency_kwargs):
            print(f"Removed: {output_dir}")
        print(f"Done removing {len(output_dirs)} decompiled output directories.")

    # def unpack_xapk(self, file_path: Path) -> Iterator[Path]:
        # TODO: Implement unpacking xapk files and place in front of decompilation when file is xapk
        # pass

    def __repr__(self) -> str:
        return f"Decompiler:(binary_paths={self.binary_paths}, extra_args={self.extra_args}, deobfuscate={self.deobfuscate}, output_suffix={self.output_suffix}, working_dir={self.working_dir}, remove_failed_output_dirs={self.remove_failed_output_dirs}, concurrent_executor={self.concurrent_executor})"



# © 2023 Lucas Faudman.
# Licensed under the MIT License (see LICENSE for details).
# For commercial use, see LICENSE for additional terms.
from yaml import safe_load as yaml_safe_load, YAMLError
from json import loads as json_loads, JSONDecodeError
from tomllib import loads as toml_loads, TOMLDecodeError
from dataclasses import dataclass, field
from typing import Optional, Iterator
from pathlib import Path
from re import (
    compile as re_compile,
    Pattern,
    IGNORECASE,
    MULTILINE,
    DOTALL,
    ASCII,
    LOCALE,
    UNICODE,
    VERBOSE,
    TEMPLATE,
)

from .concurrent_executor import ConcurrentExecutor

INCLUDED_SECRET_LOCATOR_FILES = {path.stem: path for path in (Path(__file__).parent / "secret_locators").rglob("*")}

@dataclass
class SecretLocator:
    id: str
    name: str
    pattern: Pattern
    secret_group: int | str = 0
    description: str = "No description provided."
    confidence: str = "Unknown"
    severity: str = "Unknown"
    tags: list[str] = field(default_factory=list)

    def __hash__(self) -> int:
        return hash(self.pattern)

@dataclass
class SecretResult:
    secret: bytes
    file_path: Path
    line_number: int
    locator: SecretLocator

    def __hash__(self) -> int:
        return hash(self.secret)


def try_load_json_toml_yaml(file_path: Path) -> Optional[dict]:
    if not file_path.exists():
        print(f"File not found: {file_path}. Skipping.")
        return

    with file_path.open("r") as f:
        contents = f.read()

    try:
        return json_loads(contents)
    except JSONDecodeError:
        print(f"Error loading {file_path} as JSON. Trying TOML.")

    try:
        return toml_loads(contents)
    except TOMLDecodeError:
        print(f"Error loading {file_path} as TOML. Trying YAML.")

    try:
        return yaml_safe_load(contents)
    except YAMLError:
        print(f"Error loading {file_path} as YAML. Skipping.")


def compile_str_to_bytes_pattern(pattern_str: str) -> Pattern:
    flag_notation = {
        "i": IGNORECASE,
        "m": MULTILINE,
        "s": DOTALL,
        "a": ASCII,
        "l": LOCALE,
        "u": UNICODE,
        "x": VERBOSE,
        "t": TEMPLATE,
    }

    flags = 0
    for flag_char, flag in flag_notation.items():
        if f"(?{flag_char})" in pattern_str or f"(?-{flag_char})" in pattern_str:
            flags |= flag
            pattern_str = pattern_str.replace(
                f"(?{flag_char})",
                "",
            ).replace(f"(?-{flag_char})", "")
    return re_compile(pattern_str.encode(), flags)

def load_secrets_patterns_db_format(locator_dicts: list[dict]) -> dict[str, SecretLocator]:
    secret_locators = {}
    for locator_dict in locator_dicts:
        locator_dict = locator_dict["pattern"]
        pattern_str = locator_dict.pop("regex")
        locator_dict["pattern"] = compile_str_to_bytes_pattern(pattern_str)
        locator_dict["id"] = locator_dict["name"].replace(" ", "-").lower()
        secret_locators[pattern_str] = SecretLocator(**locator_dict)

    return secret_locators

def load_gitleaks_format(locator_dicts: list[dict]) -> dict[str, SecretLocator]:
    secret_locators = {}
    for locator_dict in locator_dicts:
        pattern_str = locator_dict.pop("regex")
        locator_dict["pattern"] = compile_str_to_bytes_pattern(pattern_str)
        locator_dict["name"] = locator_dict["id"].replace("-", " ").title()
        locator_dict["secret_group"] = locator_dict.pop("secretGroup", 0)
        locator_dict["tags"] = locator_dict.pop("keywords", [])
        locator_dict.pop("entropy", None)
        locator_dict.pop("allowlist", None)
        secret_locators[pattern_str] = SecretLocator(**locator_dict)

    return secret_locators

def load_secret_locators_format(locator_dicts: list[dict]) -> dict[str, SecretLocator]:
    secret_locators = {}
    for locator_dict in locator_dicts:
        pattern_str = locator_dict.pop("pattern")
        locator_dict["pattern"] = compile_str_to_bytes_pattern(pattern_str)
        secret_locators[pattern_str] = SecretLocator(**locator_dict)
    return secret_locators

def load_simple_key_value_format(simple_locator_dict: dict) -> dict[str, SecretLocator]:
    secret_locators = {}
    for name, pattern_strs in simple_locator_dict.items():
        if isinstance(pattern_strs, str):
            pattern_strs = [pattern_strs]

        for i, pattern_str in enumerate(pattern_strs):
            locator_dict = {
                "id": name.replace(" ", "-").lower() + (f"-{i}" if i else ""),
                "name": name + (f" {i}" if i else ""),
                "pattern": compile_str_to_bytes_pattern(pattern_str),
            }
            secret_locators[pattern_str] = SecretLocator(**locator_dict)
    return secret_locators

def load_secret_locators(secret_locator_files: list[Path]) -> dict[str, SecretLocator]:
    print(f"\nLoading secret locators from {len(secret_locator_files)} files.")
    secret_locators: dict[str, SecretLocator] = {}
    for secret_locator_file in secret_locator_files:
        if not (secret_locator_file_data := try_load_json_toml_yaml(secret_locator_file)):
            continue
        if isinstance(secret_locator_file_data, list):
            secret_locators.update(load_secret_locators_format(secret_locator_file_data))
        elif locator_dicts := secret_locator_file_data.get("patterns"):
            secret_locators.update(load_secrets_patterns_db_format(locator_dicts))
        elif locator_dicts := secret_locator_file_data.get("rules"):
            secret_locators.update(load_gitleaks_format(locator_dicts))
        else:
            secret_locators.update(load_simple_key_value_format(secret_locator_file_data))

    print(f"\nLoaded {len(secret_locators)} secret locators.")
    return secret_locators

def find_secret_locator_files_by_name(secret_locator_files: list[Path]):
    existing = []
    for secret_locator_file in secret_locator_files:
        if secret_locator_file.exists():
            existing.append(secret_locator_file)
        elif secret_locator_file.stem in INCLUDED_SECRET_LOCATOR_FILES:
            existing.append(INCLUDED_SECRET_LOCATOR_FILES[secret_locator_file.stem])
    return existing

class SecretScanner:
    def __init__(
        self,
        secret_locator_files: list[Path],
        load_locators_on_init: bool = True,
        **concurrent_executor_kwargs,
    ):
        self.secret_locators: dict[str, SecretLocator] = {}
        self.load_locators_on_init = load_locators_on_init
        self.results: dict[SecretLocator, list[SecretResult]] = {}
        self.concurrent_executor = ConcurrentExecutor(**{"concurrency_type": "process", **concurrent_executor_kwargs})

        self.secret_locator_files = []
        if load_locators_on_init:
            self.load_secret_locators(secret_locator_files)

    def load_secret_locators(self, secret_locator_files: list[Path]) -> dict[str, SecretLocator]:
        secret_locator_files = find_secret_locator_files_by_name(secret_locator_files)
        self.secret_locator_files.extend(secret_locator_files)
        self.secret_locators.update(load_secret_locators(secret_locator_files))
        return self.secret_locators

    def iterscan_file(self, file_path: Path) -> Iterator[SecretResult]:
        with file_path.open("rb") as f:
            for line_number, line in enumerate(f, start=1):
                for locator in self.secret_locators.values():
                    if match := locator.pattern.search(line):
                        yield SecretResult(
                            secret=match.group(locator.secret_group),
                            file_path=file_path,
                            line_number=line_number,
                            locator=locator)

    def scan_file(self, file_path: Path) -> tuple[Path, list[SecretResult]]:
        return file_path, list(self.iterscan_file(file_path))

    def scan_concurrently(self, file_paths: Iterator[Path]) -> Iterator[tuple[Path, list[SecretResult]]]:
        yield from self.concurrent_executor.map(self.scan_file, file_paths)

    def __repr__(self) -> str:
        return f"SecretScanner(secret_locators={len(self.secret_locators)}, concurrent_executor={self.concurrent_executor}))"



# © 2023 Lucas Faudman.
# Licensed under the MIT License (see LICENSE for details).
# For commercial use, see LICENSE for additional terms.
from argparse import ArgumentParser, BooleanOptionalAction
from pathlib import Path

from .apkscan import APKScanner, Decompiler
from .secret_scanner import INCLUDED_SECRET_LOCATOR_FILES

DEFAULT_RULES = [
    INCLUDED_SECRET_LOCATOR_FILES['default']
]

BANNER_ART = """\033[1;32m   ('-.      _ (`-. .-. .-')   .-')             ('-.         .-')
  ( \033[0mOO\033[1;32m ).-. ( (\033[0mOO\033[1;32m  )\  ( \033[0mOO\033[1;32m ) ( \033[0mOO\033[1;32m ).          ( \033[0mOO\033[1;32m ).-.    ( \033[0mOO\033[1;32m )
  / . --. /_.`     \,--. ,--.(_)---\_) .-----. / . --. /,--./ ,--,
  | \-.  \(__...--''|  .'   //    _ | '  .--./ | \-.  \ |   \ |  |
  | '  |  ||  /  | ||      / \  :` `. |  |('-. | '  |  ||    \|  |
  | |_.'  ||  |_.' ||     '   '..`''. |  |\033[0mOO\033[1;32m  )| |_.'  ||  .     |
  |  .-.  ||  .___.'|  .   \ .-._)   \|  |`-'| |  .-.  ||  |\    |
  |  | |  ||  |     |  |\   \\\       /'  '--'\ |  | |  ||  | \   |
  `--' `--'`--'     `--' '--' `-----'  `-----' `--' `--'`--'  `--'   """

BANNER_TEXT = """
  \033[1;92mAPKscan v0.3.0 - \033[3mScan for secrets, endpoints, and other sensitive data
  after decompiling and deobfuscating Android files.\033[0m\033[0;92m
  (.apk, .xapk, .dex, .jar, .class, .smali, .zip, .aar, .arsc, .aab, .jadx.kts)

  \033[0m\033[2m(c) Lucas Faudman, 2024. License information in LICENSE file.
  Credits to the original authors of all dependencies used in this project.
\033[0m"""

def main():
    parser = ArgumentParser(description=BANNER_TEXT.strip())

    input_options = parser.add_argument_group("Input Options")
    input_options.add_argument(dest='files', type=Path, nargs="*", metavar="FILES_TO_SCAN",
                               help="Path(s) to Java files to decompile and scan.")
    input_options.add_argument("-r", "--rules", type=Path, nargs="*", default=DEFAULT_RULES, metavar="SECRET_LOCATOR_FILES",
                               help="Path(s) to secret locator rules/patterns files OR names of included locator sets. "
                               "\nFiles can be in SecretLocator JSON, secret-patterns-db YAML, or Gitleak TOML formats. "
                               "\nIncluded locator sets: " + ", ".join(sorted(INCLUDED_SECRET_LOCATOR_FILES.keys())) + ". "
                               + f"If not provided, default rules will be used. See: {DEFAULT_RULES[0]}"
                               )

    output_options = parser.add_argument_group("Output Options")
    output_options.add_argument("-o", "--output", type=Path, metavar="SECRETS_OUTPUT_FILE", default="secrets_output.json", help="Output file for secrets found.")
    output_options.add_argument("-f", "--format", type=str, choices=["text", "json", "yaml"], default="json", help="Output format for secrets found.")
    output_options.add_argument("-g", "--groupby", type=str, choices=["file", "locator", "both"], default="both", help="Group secrets by input file or locator. Default is 'both'.")
    output_options.add_argument("-c", "--cleanup", action=BooleanOptionalAction, default=False, help="Remove decompiled output directories after scanning.")
    output_options.add_argument("-q", "--quiet", action="store_true", help="Suppress output from subprocesses.")

    decompiler_choices = parser.add_argument_group("Decompiler Choices",
        description="Choose which decompiler(s) to use. Optionally specify path to decompiler binary. Default is JADX.")
    decompiler_choices.add_argument('--jadx', "-J", nargs='?', const=None, default=False, help="Use JADX Java decompiler.")
    decompiler_choices.add_argument('--apktool', "-A", nargs='?', const=None, default=False, help="Use APKTool SMALI disassembler.")
    decompiler_choices.add_argument('--cfr', "-C", nargs='?', const=None, default=False, help="Use CFR Java decompiler. Requires Enjarify.")
    decompiler_choices.add_argument('--procyon', "-P", nargs='?', const=None, default=False, help="Use Procyon Java decompiler. Requires Enjarify.")
    decompiler_choices.add_argument('--krakatau', "-K", nargs='?', const=None, default=False, help="Use Krakatau Java decompiler. Requires Enjarify.")
    decompiler_choices.add_argument('--fernflower', "-F", nargs='?', const=None, default=False, help="Use Fernflower Java decompiler. Requires Enjarify.")
    decompiler_choices.add_argument('--enjarify-choice', "-EC", type=str, choices=["auto", "never", "always"], default="auto", help="When to use Enjarify. Default is 'auto' which means use only when needed.")


    decompiler_options = parser.add_argument_group("Decompiler Advanced Options", description="Options for Java decompiler.")
    decompiler_options.add_argument("-d", "--deobfuscate", action=BooleanOptionalAction, default=True, help="Deobfuscate file before scanning.")
    decompiler_options.add_argument("-w", "--decompiler-working-dir", type=Path, default=Path.cwd(), help="Working directory where files will be decompiled.")
    decompiler_options.add_argument("--decompiler-output-suffix", type=str, default="-decompiled", help="Suffix for decompiled output directory names. Default is '-decompiled'.")
    decompiler_options.add_argument("--decompiler-extra-args", type=str, nargs="+", help="Additional arguments to pass to decompilers in form quoted whitespace separated '<DECOMPILER_NAME> <EXTRA_ARGS>...'. For example: --decompiler-extra-args 'jadx --no-debug-info,--no-inline'.")
    decompiler_options.add_argument("-dct", "--decompiler-concurrency-type", type=str, choices=["thread", "process", "main"], default="thread", help="Type of concurrency to use for decompilation. Default is 'thread'.")
    decompiler_options.add_argument("-dro", "--decompiler-results-order", type=str, choices=["completed", "submitted"], default="completed", help="Order to process results from decompiler. Default is 'completed'.")
    decompiler_options.add_argument("-dmw", "--decompiler-max-workers", type=int, default=None, help="Maximum number of workers to use for decompilation.")
    decompiler_options.add_argument("-dcs", "--decompiler-chunksize", type=int, default=1, help="Number of files to decompile per thread/process.")
    decompiler_options.add_argument("-dto", "--decompiler-timeout", type=int, help="Timeout for decompilation in seconds.")

    scanner_options = parser.add_argument_group("Secret Scanner Advanced Options", description="Options for secret scanner.")
    scanner_options.add_argument("-sct", "--scanner-concurrency-type", type=str, choices=["thread", "process", "main"], default="process", help="Type of concurrency to use for scanning. Default is 'process'.")
    scanner_options.add_argument("-sro", "--scanner-results-order", type=str, choices=["completed", "submitted"], default="completed", help="Order to process results from scanner. Default is 'completed'.")
    scanner_options.add_argument("-smw", "--scanner-max-workers", type=int, default=None, help="Maximum number of workers to use for scanning.")
    scanner_options.add_argument("-scs", "--scanner-chunksize", type=int, default=1, help="Number of files to scan per thread/process.")
    scanner_options.add_argument("-sto", "--scanner-timeout", type=int, help="Timeout for scanning in seconds.")

    args = parser.parse_args()
    if not args.quiet:
        print(BANNER_ART + BANNER_TEXT)
        print('\033[1mStarting APKscan...\033[0m\n')

    if not args.files:
        parser.error("No files to scan provided. Please provide at least one file to scan.")

    decompiler_kwargs = {
        "binaries": {},
        "enjarify_choice": args.enjarify_choice,
        "deobfuscate": args.deobfuscate,
        "remove_failed_output_dirs": args.cleanup,
        "suppress_output": args.quiet,
    }
    scanner_kwargs = {
        "secret_locator_files": args.rules,
    }
    for k, v in vars(args).items():
        if k.startswith("scanner_"):
            scanner_kwargs[k[8:]] = v
        elif k.startswith("decompiler_"):
            decompiler_kwargs[k[11:]] = v
        elif k in Decompiler.CONFIG and v is not False:
            decompiler_kwargs['binaries'][k] = v

    apk_scanner = APKScanner(
        decompiler_kwargs=decompiler_kwargs,
        scanner_kwargs=scanner_kwargs,
        output_file=args.output,
        output_format=args.format,
        groupby=args.groupby,
        cleanup=args.cleanup
    )

    try:
        apk_scanner.decompile_and_scan(args.files)
    except KeyboardInterrupt:
        print("\n\033[1;34mWriting output, cleaning up, and exiting...\033[0m")
    finally:
        apk_scanner.write_output()
        apk_scanner.do_cleanup()

    if apk_scanner.secrets_results:
        print(f"\033[1;32m\nAPKscan done. Secrets saved to {apk_scanner.output_file}\033[0m")
        exit(0)
    else:
        print("\033[1;32m\nAPKscan done. \033[1;31mNo secrets found.\033[0m")
        exit(1)

if __name__ == "__main__":
    main()



# © 2023 Lucas Faudman.
# Licensed under the MIT License (see LICENSE for details).
# For commercial use, see LICENSE for additional terms.
from setuptools import setup, find_namespace_packages
import setuptools_scm

setup(
    name='apkscan',
    version='0.3.0',
    use_scm_version=True,
    setup_requires=['setuptools_scm>=8'],
    description='Scan for secrets, endpoints, and other sensitive data after decompiling and deobfuscating Android files. (.apk, .xapk, .dex, .jar, .class, .smali, .zip, .aar, .arsc, .aab, .jadx.kts)',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Lucas Faudman',
    author_email='lucasfaudman@gmail.com',
    url='https://github.com/LucasFaudman/apkscan.git',
    packages=find_namespace_packages(where='src', exclude=['tests*']),
    package_dir={'': 'src'},
    package_data={
        '': ['LICENSE'],
        'apkscan.secret_locators': ['*.json', '*.yaml', '*.yml', '*.toml'],
    },
    include_package_data=True,
    exclude_package_data={'': ['.gitignore', '.pre-commit-config.yaml']},
    install_requires=[
        'enjarify-adapter',
        'pyyaml',
    ],
    entry_points={
        'console_scripts': [
            'apkscan = apkscan.main:main',
        ],
    },
    python_requires='>=3.11',
    license='LICENSE',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Security',
        'Topic :: Utilities',
    ],
    keywords='secret scanner jadx decompile android java security mobile decompiler concurrency penetration-testing apktool security-tools fernflower cfr jadx procyon enjarify krakatau secret-scanning decompiler-java secret-scanner apk dex jar class smali zip aar arsc aab xapk jadx.kts ',
    project_urls={
        'Homepage': 'https://github.com/LucasFaudman/apkscan.git',
        'Repository': 'https://github.com/LucasFaudman/apkscan.git',
    },
)



# © 2023 Lucas Faudman.
# Licensed under the MIT License (see LICENSE for details).
# For commercial use, see LICENSE for additional terms.
from .decompiler import Decompiler
from .secret_scanner import SecretScanner, SecretLocator, SecretResult, load_secret_locators
from .apkscan import APKScanner
